import fnmatch
import logging

import AppKit
from ApplicationServices import (
    AXIsProcessTrusted,
    AXUIElementCreateApplication,
    AXUIElementCreateSystemWide,
    CFEqual,
)
from macuiauto import _converter
from macuiauto._macos import (
    PAXUIElementCopyActionNames,
    PAXUIElementCopyAttributeNames,
    PAXUIElementCopyAttributeValue,
    PAXUIElementCopyElementAtPosition,
    PAXUIElementGetPid,
    PAXUIElementIsAttributeSettable,
    PAXUIElementPerformAction,
    PAXUIElementSetAttributeValue,
    PAXUIElementSetMessagingTimeout,
)
from macuiauto.errors import (
    AXError,
    AXErrorAPIDisabled,
    AXErrorCannotComplete,
    AXErrorIllegalArgument,
    AXErrorNotImplemented,
    AXErrorNoValue,
    AXErrorUnsupported,
)
from PyObjCTools import AppHelper

logger = logging.getLogger(__name__)


class RunningApplication(object):
    def __init__(self, app_ref):
        self.ref = app_ref

    def process_identifier(self):
        return self.ref.processIdentifier()

    def terminate(self):
        return self.ref.terminate()

    def localized_name(self):
        return self.ref.localizedName()

    def bundle_identifier(self):
        return self.ref.bundleIdentifier()

    def activate_with_options(self, options: int):
        return self.ref.activateWithOptions_(options)


class AXUIElement(object):
    def __init__(self, ref=None):
        self.ref = ref
        self.converter = _converter.Converter(self.__class__)

    def __repr__(self):
        c = repr(self.__class__).partition("<class '")[-1].rpartition("'>")[0]

        for element_describer in ("AXTitle", "AXValue", "AXRoleDescription"):
            title = str(getattr(self, element_describer, ""))
            if title:
                break

        role = getattr(self, "AXRole", "<No role!>")

        return "<%s %s %s>" % (c, role, title)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.ref is None and other.ref is None:
            return True

        if self.ref is None or other.ref is None:
            return False

        return CFEqual(self.ref, other.ref)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, item):
        if item in self._ax_attributes:
            return self._get_ax_attribute(item)
        elif item in self._ax_actions:

            def perform_ax_action():
                self.perform_ax_action(item)

            return perform_ax_action
        else:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self), item)
            )

    def __setattr__(self, key, value):
        if key.startswith("AX"):
            try:
                if key in self._ax_attributes:
                    self._set_ax_attribute(key, value)
            except AXErrorIllegalArgument:
                pass
        else:
            super(AXUIElement, self).__setattr__(key, value)

    def __dir__(self):
        return (
            self._ax_attributes
            + self._ax_actions
            + list(self.__dict__.keys())
            + dir(super(AXUIElement, self))  # not working in python 2
        )

    @classmethod
    def _from_bundle_id(cls, bundle_id) -> 'AXUIElement':
        """
        Get the top level element for the application with the specified
        bundle ID, such as com.vmware.fusion.
        """
        apps = running_apps_with_bundle_id(bundle_id)
        if not apps:
            raise ValueError(
                "Specified bundle ID not found in " "running apps: %s" % bundle_id
            )
        return cls._from_pid(apps[0].process_identifier())

    @classmethod
    def _from_localized_name(cls, name) -> 'AXUIElement':
        """Get the top level element for the application with the specified
        localized name, such as VMware Fusion.

        Wildcards are also allowed.
        """
        # Refresh the runningApplications list
        apps = get_running_apps()
        for app in apps:
            if fnmatch.fnmatch(app.localized_name(), name):
                pid = app.process_identifier()
                return cls._from_pid(pid)
        raise ValueError("Specified application not found in running apps.")

    @classmethod
    def _from_pid(cls, pid: int) -> 'AXUIElement':
        """
        Creates an instance with the AXUIElementRef for the application with
        the specified process ID.
        """
        app_ref = AXUIElementCreateApplication(pid)

        return cls(ref=app_ref)

    @classmethod
    def _frontmost(cls) -> 'AXUIElement':
        """
        Creates an instance with the AXUIElementRef for the frontmost application.
        """
        for app in get_running_apps():
            pid = app.process_identifier()
            ref = cls._from_pid(pid)
            try:
                if ref.AXFrontmost:
                    return ref
            except (
                AttributeError,
                AXErrorUnsupported,
                AXErrorCannotComplete,
                AXErrorAPIDisabled,
                AXErrorNotImplemented,
            ):
                # Some applications do not have an explicit GUI
                # and so will not have an AXFrontmost attribute
                # Trying to read attributes from Google Chrome Helper returns
                # ErrorAPIDisabled for some reason - opened radar bug 12837995
                pass
        raise ValueError("No GUI application found.")

    @classmethod
    def _systemwide(cls) -> 'AXUIElement':
        """
        Creates an instance with the AXUIElementRef for the system-wide
        accessibility object.
        """
        app_ref = AXUIElementCreateSystemWide()
        return cls(ref=app_ref)

    @classmethod
    def _with_window(cls) -> 'AXUIElement':
        """
        Creates an instance with the AXUIElementRef for a random application
        that has windows.
        """
        for app in get_running_apps():
            pid = app.process_identifier()
            ref = cls._from_pid(pid)
            if hasattr(ref, "windows") and len(ref.windows()) > 0:
                return ref
        raise ValueError("No GUI application found.")

    @property
    def _ax_actions(self) -> list[str]:
        """Gets the list of actions available on the AXUIElement"""
        try:
            names = PAXUIElementCopyActionNames(self.ref)
            return list(names)
        except AXError:
            return []

    @property
    def _ax_attributes(self) -> list[str]:
        """Gets the list of attributes available on the AXUIElement"""
        try:
            names = PAXUIElementCopyAttributeNames(self.ref)
            return list(names)
        except AXError:
            return []

    @property
    def bundle_id(self) -> str:
        """Gets the AXUIElement's bundle identifier"""
        return self._running_app.bundle_identifier()

    @property
    def pid(self) -> int:
        """Gets the AXUIElement's process ID"""
        pid = PAXUIElementGetPid(self.ref)
        return pid

    @property
    def _running_app(self) -> RunningApplication:
        ra = AppKit.NSRunningApplication
        app = ra.runningApplicationWithProcessIdentifier_(self.pid)
        return RunningApplication(app)

    def _get_element_at_position(self, coord: tuple) -> 'AXUIElement':
        if self.ref is None:
            raise AXErrorUnsupported(
                "Operation not supported on null element references"
            )
        element = PAXUIElementCopyElementAtPosition(self.ref, float(coord[0]), float(coord[1]))

        return self.__class__(element)

    def _set_timeout(self, timeout):
        """
        Sets the timeout value used in the accessibility API

        Args:
            timeout: timeout in seconds
        """
        PAXUIElementSetMessagingTimeout(self.ref, timeout)

    def _activate(self):
        """Activates the application (bringing menus and windows forward)"""
        # NSApplicationActivateAllWindows | NSApplicationActivateIgnoringOtherApps
        # == 3 - PyObjC in 10.6 does not expose these constants though so I have
        # to use the int instead of the symbolic names
        self._running_app.activate_with_options(3)

    def _get_ax_attribute(self, item: str) -> any:
        """Gets the value of the the specified attribute"""
        if item in self._ax_attributes:
            try:
                attr_value = PAXUIElementCopyAttributeValue(self.ref, item)
                return self.converter.convert_value(attr_value)
            except AXErrorNoValue:
                if item == "AXChildren":
                    return []
                return None

        raise AttributeError("'%s' object has no attribute '%s'" % (type(self), item))

    def _set_ax_attribute(self, name: str, value: any):
        """Sets the specified attribute to the specified value"""
        settable = PAXUIElementIsAttributeSettable(self.ref, name)

        if not settable:
            raise AXErrorUnsupported("Attribute is not settable")

        PAXUIElementSetAttributeValue(self.ref, name, value)

    def perform_ax_action(self, name: str):
        """Performs specified action on the AXUIElementRef"""
        PAXUIElementPerformAction(self.ref, name)


def axenabled() -> bool:
    """Return the status of accessibility on the system"""
    return AXIsProcessTrusted()


def get_frontmost_pid() -> int:
    """Return the process ID of the application in the foreground"""
    frontmost_app = AppKit.NSWorkspace.sharedWorkspace().frontmostApplication()
    pid = frontmost_app.process_identifier()
    return pid


def get_running_apps() -> list['RunningApplication']:
    """Get a list of the running applications"""
    AppHelper.callLater(1, AppHelper.stopEventLoop)
    AppHelper.runConsoleEventLoop()
    # Get a list of running applications
    ws = AppKit.NSWorkspace.sharedWorkspace()
    apps = ws.runningApplications()
    app_list = []
    for app in apps:
        app_list.append(RunningApplication(app))
    return app_list


def launch_app_by_bundle_id(bundle_id: str):
    # NSWorkspaceLaunchAllowingClassicStartup does nothing on any
    # modern system that doesn't have the classic environment installed.
    # Encountered a bug when passing 0 for no options on 10.6 PyObjC.
    ws = AppKit.NSWorkspace.sharedWorkspace()

    r = ws.launchAppWithBundleIdentifier_options_additionalEventParamDescriptor_launchIdentifier_(  # noqa: B950
        bundle_id,
        AppKit.NSWorkspaceLaunchAllowingClassicStartup,
        AppKit.NSAppleEventDescriptor.nullDescriptor(),
        None,
    )
    # On 10.6, this returns a tuple - first element bool result, second is
    # a number. Let's use the bool result.
    if not r[0]:
        raise RuntimeError("Error launching specified application. %s" % str(r))


def launch_app_by_bundle_path(bundle_path: str, arguments=None):
    if arguments is None:
        arguments = []

    bundleUrl = AppKit.NSURL.fileURLWithPath_(bundle_path)
    workspace = AppKit.NSWorkspace.sharedWorkspace()
    configuration = {AppKit.NSWorkspaceLaunchConfigurationArguments: arguments}

    return workspace.launchApplicationAtURL_options_configuration_error_(
        bundleUrl, AppKit.NSWorkspaceLaunchAllowingClassicStartup, configuration, None
    )


def terminate_app_by_bundle_id(bundle_id: str):
    apps = running_apps_with_bundle_id(bundle_id)

    if not apps:
        return False

    return apps[0].terminate()


def terminate_app_by_pid(pid: int):
    apps = get_running_apps()

    for app in apps:
        if pid == app.process_identifier():
            return app.terminate()
    return False


def running_apps_with_bundle_id(bundle_id: str) -> list['RunningApplication']:
    """
    Returns an array of NSRunningApplications, or an empty array if
    no applications match the bundle identifier.
    """
    ra = AppKit.NSRunningApplication
    apps = ra.runningApplicationsWithBundleIdentifier_(bundle_id)
    app_list = []
    for app in apps:
        app_list.append(RunningApplication(app))
    return app_list




import time
from collections import deque
from macuiauto import AXCallbacks
from macuiauto import errors
from macuiauto import _a11y
from macuiauto._converter import *
from macuiauto._mixin import KeyboardMouseMixin, WaitForMixin
import warnings
from typing import Optional, List, Any, Callable, Dict

class NativeUIElement(
    KeyboardMouseMixin, WaitForMixin, _a11y.AXUIElement
):
    """NativeUIElement class - expose the accessibility API in the simplest,
    most natural way possible.
    """

    def __init__(self, ref=None):
        self._last_method: Optional[Callable[..., Optional[Any]]] = None
        self._last_kwargs: Optional[Dict[str, Any]] = None
        self._last_result: Optional[Any] = None
        super(NativeUIElement, self).__init__(ref=ref)
        self.eventList = deque()

    def __getattr__(self, name):
        """Handle attribute requests in several ways:

        1. If it starts with AX, it is probably an a11y attribute. Pass
           it to the handler in _a11y which will determine that for sure.
        2. See if the attribute is an action which can be invoked on the
           UIElement. If so, return a function that will invoke the attribute.
        """
        print(name)
        if name not in ['find_first_r', 'find_first']:
            self._last_method = None
            self._last_kwargs = None
            self._last_result = None

        if "AX" + name in self._ax_actions:
            action = super(NativeUIElement, self).__getattr__("AX" + name)

            def perform_specified_action():
                # activate the app before performing the specified action
                self._activate()
                return action()

            return perform_specified_action
        else:
            return super(NativeUIElement, self).__getattr__(name)

    @property
    def title(self) -> str:
        return self.__getattr__('AXTitle')

    @title.setter
    def title(self, new_value):
        self.__setattr__('AXTitle', new_value)

    @property
    def frame(self) -> CGRect:
        return self.__getattr__('AXFrame')

    @property
    def position(self) -> CGPoint:
        return self.__getattr__('AXPosition')

    @property
    def role(self) -> str:
        return self.__getattr__('AXRole')

    @property
    def parent(self) -> Optional['NativeUIElement']:
        return self.__getattr__('AXParent')

    @property
    def size(self) -> CGSize:
        return self.__getattr__('AXSize')

    @property
    def children(self) -> list['NativeUIElement']:
        return self.__getattr__('AXChildren')

    @property
    def enabled(self) -> bool:
        return self.__getattr__('AXEnabled')

    @property
    def description(self) -> str:
        return self.__getattr__('AXDescription')

    @property
    def label(self) -> str:
        warnings.warn(
            f"Please use description instead",
            DeprecationWarning,
            stacklevel=2
        )
        return self.__getattr__('AXDescription')


    @property
    def focused(self) -> bool:
        return self.__getattr__('AXFocused')

    @property
    def value(self) -> str:
        return self.__getattr__('AXValue')

    @value.setter
    def value(self, new_value):
        self.__setattr__('AXValue', new_value)

    @property
    def role_description(self) -> str:
        return self.__getattr__('AXRoleDescription')

    @property
    def window(self) -> Optional['NativeUIElement']:
        return self.__getattr__('AXWindow')

    def press(self):
        if "Press" in self._ax_actions:
            return super(NativeUIElement, self).perform_ax_action("Press")

    def show_menu(self):
        if "ShowMenu" in self._ax_actions:
            return super(NativeUIElement, self).perform_ax_action("ShowMenu")

    @classmethod
    def get_app_ref_by_pid(cls, pid) -> Optional['NativeUIElement']:
        """Get the top level element for the application specified by pid."""
        return cls._from_pid(pid)

    @classmethod
    def get_app_ref_by_bundle_id(cls, bundle_id: str) -> Optional['NativeUIElement']:
        """
        Get the top level element for the application with the specified
        bundle ID, such as com.vmware.fusion.
        """
        return cls._from_bundle_id(bundle_id)

    @classmethod
    def get_app_ref_by_localized_name(cls, name: str) -> Optional['NativeUIElement']:

        """Get the top level element for the application with the specified
        localized name, such as VMware Fusion.

        Wildcards are also allowed.
        """
        # Refresh the runningApplications list
        return cls._from_localized_name(name)

    @classmethod
    def get_frontmost_app(cls) -> Optional['NativeUIElement']:
        """Get the current frontmost application.

        Raise a ValueError exception if no GUI applications are found.
        """
        # Refresh the runningApplications list
        return cls._frontmost()

    @classmethod
    def get_any_app_with_window(cls) -> Optional['NativeUIElement']:
        """Get a random app that has windows.

        Raise a ValueError exception if no GUI applications are found.
        """
        # Refresh the runningApplications list
        return cls._with_window()

    @classmethod
    def get_system_object(cls) -> Optional['NativeUIElement']:
        """Get the top level system accessibility object."""
        return cls._systemwide()

    @classmethod
    def set_system_wide_timeout(cls, timeout=0.0):
        """Set the system-wide accessibility timeout.

        Args:
            timeout: non-negative float. 0 will reset to the system default.

        Returns:
            None

        """
        return cls.set_systemwide_timeout(timeout)

    @staticmethod
    def launch_app_by_bundle_id(bundle_id: str):
        """Launch the application with the specified bundle ID"""
        # NSWorkspaceLaunchAllowingClassicStartup does nothing on any
        # modern system that doesn't have the classic environment installed.
        # Encountered a bug when passing 0 for no options on 10.6 PyObjC.
        _a11y.launch_app_by_bundle_id(bundle_id)

    @staticmethod
    def launch_app_by_bundle_path(bundle_path: str, arguments=None):
        """Launch app with a given bundle path.

        Return True if succeed.
        """
        return _a11y.launch_app_by_bundle_path(bundle_path, arguments)

    @staticmethod
    def terminate_app_by_bundle_id(bundle_id: str):
        """Terminate app with a given bundle ID.
        Requires 10.6.

        Return True if succeed.
        """
        return _a11y.terminate_app_by_bundle_id(bundle_id)

    @classmethod
    def set_systemwide_timeout(cls, timeout=0.0):
        """Set the system-wide accessibility timeout.

        Args:
            timeout: A value of 0 will reset the timeout to the system default.
        """
        return cls.systemwide().setTimeout(timeout)

    def set_timeout(self, timeout=0.0):
        """Set the accessibiltiy API timeout on the given reference.

        Args:
            timeout: A value of 0 will reset the timeout to the systemwide value
        """
        self._set_timeout(timeout)

    def get_attributes(self) -> list[str]:
        """Get a list of the attributes available on the element."""
        return self._ax_attributes

    def get_actions(self) -> list[str]:
        """Return a list of the actions available on the element."""
        actions = self._ax_actions
        # strip leading AX from actions - help distinguish them from attributes
        return [action[2:] for action in actions]

    def set_string(self, attribute, string):
        """Set the specified attribute to the specified string."""
        return self.__setattr__(attribute, str(string))

    def get_element_at_position(self, coord: tuple) -> Optional['NativeUIElement']:
        """Return the AXUIElement at the given coordinates.

        If self is behind other windows, this function will return self.
        """
        return self._get_element_at_position(coord)

    def activate(self):
        """Activate the application (bringing menus and windows forward)"""
        return self._activate()

    def get_application(self) -> Optional['NativeUIElement']:
        """Get the base application UIElement.

        If the UIElement is a child of the application, it will try
        to get the AXParent until it reaches the top application level
        element.
        """
        app = self
        while "AXParent" in app._ax_attributes:
            app = app.AXParent
        return app

    def menu_item(self, *args) -> Optional['NativeUIElement']:
        """Return the specified menu item.

        Example - refer to items by name:

        app.menuItem('File', 'New').Press()
        app.menuItem('Edit', 'Insert', 'Line Break').Press()

        Refer to items by index:

        app.menuitem(1, 0).Press()

        Refer to items by mix-n-match:

        app.menuitem(1, 'About TextEdit').Press()
        """
        menuitem = self.get_application().AXMenuBar
        return self._menu_item(menuitem, *args)

    def pop_up_item(self, *args) -> Optional['NativeUIElement']:
        """Return the specified item in a pop up menu."""
        self.Press()
        time.sleep(0.5)
        return self._menu_item(self, *args)

    def get_bundle_id(self) -> str:
        """Return the bundle ID of the application."""
        return self.bundle_id

    def get_localized_name(self) -> str:
        """Return the localized name of the application."""
        return self.get_application().AXTitle

    '''
    Search 
    '''
    def _generate_children(self, target=None, recursive=False) -> List['NativeUIElement']:
        """Generator which yields all AXChildren of the object."""
        if target is None:
            target = self

        try:
            children = target.children
        except (AttributeError, errors.AXError):
            return

        for child in children:
            yield child
            if recursive:
                for c in self._generate_children(child, recursive):
                    yield c

    def _find_all(self, recursive=False, **kwargs) -> List['NativeUIElement']:
        """Return a list of all children that match the specified criteria."""
        return filter(
            AXCallbacks.match_filter(**kwargs),
            self._generate_children(recursive=recursive),
        )

    def _find_first(self, recursive=False, **kwargs) -> Optional['NativeUIElement']:
        """Return the first object that matches the criteria."""
        for item in self._find_all(recursive=recursive, **kwargs):
            return item

    def find_first(self, **kwargs) -> Optional['NativeUIElement']:
        """Return the first object that matches the criteria."""
        return self._find_first(**kwargs)

    def find_first_r(self, **kwargs) -> Optional['NativeUIElement']:
        """Search recursively for the first object that matches the
        criteria.
        """
        self._last_method = self._find_first
        self._last_kwargs = kwargs
        self._last_result = self._find_first(recursive=True, **kwargs)
        return self._last_result

    def find_all(self, **kwargs) -> List['NativeUIElement']:
        """Return a list of all children that match the specified criteria."""
        return list(self._find_all(**kwargs))

    def find_all_r(self, **kwargs) -> List['NativeUIElement']:
        """Return a list of all children (recursively) that match
        the specified criteria.
        """
        return list(self._find_all(recursive=True, **kwargs))

    def _convenience_match(self, role, attr, match) -> List['NativeUIElement']:
        """Method used by role based convenience functions to find a match"""
        kwargs = {}
        # If the user supplied some text to search for,
        # supply that in the kwargs
        if match:
            kwargs[attr] = match
        return self.find_all(AXRole=role, **kwargs)

    def _convenience_match_r(self, role, attr, match) -> List['NativeUIElement']:
        """Method used by role based convenience functions to find a match"""
        kwargs = {}
        # If the user supplied some text to search for,
        # supply that in the kwargs
        if match:
            kwargs[attr] = match
        return self.find_all_r(AXRole=role, **kwargs)

    def text_areas(self, match=None) -> List['NativeUIElement']:
        """Return a list of text areas with an optional match parameter."""
        return self._convenience_match("AXTextArea", "AXTitle", match)

    def text_areas_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of text areas with an optional match parameter."""
        return self._convenience_match_r("AXTextArea", "AXTitle", match)

    def text_fields(self, match=None) -> List['NativeUIElement']:
        """Return a list of textfields with an optional match parameter."""
        return self._convenience_match("AXTextField", "AXRoleDescription", match)

    def text_fields_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of textfields with an optional match parameter."""
        return self._convenience_match_r("AXTextField", "AXRoleDescription", match)

    def buttons(self, match=None) -> List['NativeUIElement']:
        """Return a list of buttons with an optional match parameter."""
        return self._convenience_match("AXButton", "AXTitle", match)

    def buttons_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of buttons with an optional match parameter."""
        return self._convenience_match_r("AXButton", "AXTitle", match)

    def windows(self, match=None) -> list['NativeUIElement']:
        """Return a list of windows with an optional match parameter."""
        return self._convenience_match("AXWindow", "AXTitle", match)

    def windows_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of windows with an optional match parameter."""
        return self._convenience_match_r("AXWindow", "AXTitle", match)

    def sheets(self, match=None) -> List['NativeUIElement']:
        """Return a list of sheets with an optional match parameter."""
        return self._convenience_match("AXSheet", "AXDescription", match)

    def sheets_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of sheets with an optional match parameter."""
        return self._convenience_match_r("AXSheet", "AXDescription", match)

    def static_texts(self, match=None) -> List['NativeUIElement']:
        """Return a list of statictexts with an optional match parameter."""
        return self._convenience_match("AXStaticText", "AXValue", match)

    def static_texts_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of statictexts with an optional match parameter"""
        return self._convenience_match_r("AXStaticText", "AXValue", match)

    def generic_elements(self, match=None) -> List['NativeUIElement']:
        """Return a list of genericelements with an optional match parameter."""
        return self._convenience_match("AXGenericElement", "AXValue", match)

    def generic_elements_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of genericelements with an optional match parameter."""
        return self._convenience_match_r("AXGenericElement", "AXValue", match)

    def groups(self, match=None) -> List['NativeUIElement']:
        """Return a list of groups with an optional match parameter."""
        return self._convenience_match("AXGroup", "AXRoleDescription", match)

    def groups_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of groups with an optional match parameter."""
        return self._convenience_match_r("AXGroup", "AXRoleDescription", match)

    def radio_buttons(self, match=None) -> List['NativeUIElement']:
        """Return a list of radio buttons with an optional match parameter."""
        return self._convenience_match("AXRadioButton", "AXTitle", match)

    def radio_buttons_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of radio buttons with an optional match parameter."""
        return self._convenience_match_r("AXRadioButton", "AXTitle", match)

    def pop_up_buttons(self, match=None) -> List['NativeUIElement']:
        """Return a list of popup menus with an optional match parameter."""
        return self._convenience_match("AXPopUpButton", "AXTitle", match)

    def pop_up_buttons_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of popup menus with an optional match parameter."""
        return self._convenience_match_r("AXPopUpButton", "AXTitle", match)

    def rows(self, match=None) -> List['NativeUIElement']:
        """Return a list of rows with an optional match parameter."""
        return self._convenience_match("AXRow", "AXTitle", match)

    def rows_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of rows with an optional match parameter."""
        return self._convenience_match_r("AXRow", "AXTitle", match)

    def sliders(self, match=None) -> List['NativeUIElement']:
        """Return a list of sliders with an optional match parameter."""
        return self._convenience_match("AXSlider", "AXValue", match)

    def sliders_r(self, match=None) -> List['NativeUIElement']:
        """Return a list of sliders with an optional match parameter."""
        return self._convenience_match_r("AXSlider", "AXValue", match)

    def _menu_item(self, menuitem, *args) -> Optional['NativeUIElement']:
        """Return the specified menu item.

        Example - refer to items by name:

        app._menuItem(app.AXMenuBar, 'File', 'New').Press()
        app._menuItem(app.AXMenuBar, 'Edit', 'Insert', 'Line Break').Press()

        Refer to items by index:

        app._menuitem(app.AXMenuBar, 1, 0).Press()

        Refer to items by mix-n-match:

        app._menuitem(app.AXMenuBar, 1, 'About TextEdit').Press()
        """
        self._activate()
        for item in args:
            # If the item has an AXMenu as a child, navigate into it.
            # This seems like a silly abstraction added by apple's a11y api.
            if menuitem.AXChildren[0].AXRole == "AXMenu":
                menuitem = menuitem.AXChildren[0]
            # Find AXMenuBarItems and AXMenuItems using a handy wildcard
            try:
                menuitem = menuitem.AXChildren[int(item)]
            except ValueError:
                menuitem = menuitem.findFirst(AXRole="AXMenu*Item", AXTitle=item)
        return menuitem

    def wait(self, timeout: int = 30, poll_interval: float = 0.1) -> Optional[Any]:
        """Wait for the last called method to return a non-None result.

        Args:
            timeout (int): Maximum time to wait in seconds.
            poll_interval (float): Time interval between method calls in seconds.

        Returns:
            Optional[Any]: The result of the method call if found within the timeout, otherwise None.
        """
        if self._last_method is None:
            raise ValueError("No method has been called to wait for.")

        start_time = time.time()
        while (time.time() - start_time) < timeout:
            self._last_result = self._last_method(**self._last_kwargs)
            if self._last_result is not None:
                return self._last_result
            time.sleep(poll_interval)
        return None

    def all_values(self):
        print(self.get_attributes())
        for key in self._ax_attributes:
            print(key, self.__getattr__(key))



"""Automated Testing on macOS"""
# flake8: noqa: F401
__version__ = "3.2.0"

from macuiauto import _a11y, errors
from macuiauto.AXClasses import NativeUIElement
from pynput import keyboard

Key = keyboard.Key
Error = errors.AXError
ErrorAPIDisabled = errors.AXErrorAPIDisabled
ErrorInvalidUIElement = errors.AXErrorInvalidUIElement
ErrorCannotComplete = errors.AXErrorCannotComplete
ErrorUnsupported = errors.AXErrorUnsupported
ErrorNotImplemented = errors.AXErrorNotImplemented

#
# get_app_ref_by_localized_name = NativeUIElement.getAppRefByLocalizedName
# terminate_app_by_bundle_id = NativeUIElement.terminateAppByBundleId
# launch_app_by_bundle_path = NativeUIElement.launchAppByBundlePath
# set_system_wide_timeout = NativeUIElement.setSystemWideTimeout
# get_app_ref_by_bundle_id = NativeUIElement.getAppRefByBundleId
# launch_app_by_bundle_id = NativeUIElement.launchAppByBundleId
# get_frontmost_app = NativeUIElement.getFrontmostApp
get_app_ref_by_pid = NativeUIElement.get_app_ref_by_pid
get_app_ref_by_bundle_id = NativeUIElement.get_app_ref_by_bundle_id
get_running_apps = _a11y.get_running_apps
launch_app_by_bundle_id = _a11y.launch_app_by_bundle_id
launch_app_by_bundle_path = _a11y.launch_app_by_bundle_path
terminate_app_by_bundle_id = _a11y.terminate_app_by_bundle_id
terminate_app_by_pid = _a11y.terminate_app_by_pid
running_apps_with_bundle_id = _a11y.running_apps_with_bundle_id





from pynput.mouse import Button
from pynput.keyboard import Key
from pynput import mouse
from pynput import keyboard


class Mouse(object):
    mouse = mouse.Controller()
    keyboard = keyboard.Controller()


    def drag_mouse_button_left(self, coord: tuple, dest_coord: tuple):
        """Drag the left mouse button without modifiers pressed.

        Parameters: coordinates to click on screen (tuple (x, y))
                    dest coordinates to drag to (tuple (x, y))
                    interval to send event of btn down, drag and up
        Returns: None
        """
        self.mouse.position = coord
        self.mouse.press(Button.left)
        self.mouse.position = dest_coord
        self.mouse.release(Button.left)

    def double_click_drag_mouse_button_left(self, coord: tuple, dest_coord: tuple):
        """Double-click and drag the left mouse button without modifiers
        pressed.

        Parameters: coordinates to double-click on screen (tuple (x, y))
                    dest coordinates to drag to (tuple (x, y))
                    interval to send event of btn down, drag and up
        Returns: None
        """
        self.mouse.position = coord
        self.mouse.press(Button.left)
        self.mouse.release(Button.left)
        self.mouse.press(Button.left)
        self.mouse.position = dest_coord

    def click_mouse_button_left(self, coord=None, count=1):
        """Click the left mouse button without modifiers pressed.

        Parameters: coordinates to click on screen (tuple (x, y))
        Returns: None
        """
        if coord is not None:
            self.mouse.position = coord
        self.mouse.click(Button.left, count)

    def click_mouse_button_right(self, coord=None, count=1):
        """Click the right mouse button without modifiers pressed.

        Parameters: coordinates to click on scren (tuple (x, y))
        Returns: None
        """
        if coord is not None:
            self.mouse.position = coord
        self.mouse.click(Button.right, count)

    def click_mouse_button_left_with_modifiers(self, coord: tuple, modifiers: list[Key], count=1):
        """Click the left mouse button with modifiers pressed.

        Parameters: coordinates to click; modifiers (list) (e.g. ["shift"] or
                    ["command", "shift"])
        Returns: None
        """
        with self.keyboard.pressed(**modifiers):
            self.click_mouse_button_left(coord, count)

    def click_mouse_button_right_with_modifiers(self, coord: tuple, modifiers: list[Key], count=1):
        """Click the right mouse button with modifiers pressed.

        Parameters: coordinates to click; modifiers (list)
        Returns: None
        """
        with self.keyboard.pressed(**modifiers):
            self.click_mouse_button_right(coord, count)

    def double_click_mouse_left(self, coord: tuple):
        """Double-click primary mouse button.

        Parameters: coordinates to click (assume primary is left button)
        Returns: None
        """
        self.mouse.position = coord
        self.mouse.click(Button.left, 2)

    def double_mouse_button_left_with_modifiers(self, coord: tuple, modifiers: list[Key]):
        """Click the left mouse button with modifiers pressed.

        Parameters: coordinates to click; modifiers (list)
        Returns: None
        """
        self.mouse.position = coord
        with self.keyboard.pressed(**modifiers):
            self.mouse.click(Button.left, 2)

    def triple_click_mouse_left(self, coord: tuple):
        """Triple-click primary mouse button.

        Parameters: coordinates to click (assume primary is left button)
        Returns: None
        """
        self.mouse.position = coord
        self.mouse.click(Button.left, 3)


class Keyboard(object):
    keyboard = keyboard.Controller()

    def send_key(self, keychr: str | Key):
        """Send one character with no modifiers."""
        self.keyboard.press(keychr)
        self.keyboard.release(keychr)

    def send_key_with_modifiers(self, keychr: str, modifiers: list[Key]):
        """Send one character with modifiers pressed

        Parameters: key character, modifiers (list) (e.g. ["shift"] or
                    ["command", "shift"]
        """
        with self.keyboard.pressed(*modifiers):
            self.keyboard.press(keychr)
            self.keyboard.release(keychr)

    def type(self, keystr: str):
        """Send a series of characters with no modifiers."""
        self.keyboard.type(keystr)


class KeyboardMouseMixin(Mouse, Keyboard):
    pass



from macuiauto import AXCallbacks
from macuiauto._notification import Observer


class WaitForMixin(object):

    def wait_for(self, timeout, notification, **kwargs):
        """Generic wait for a UI event that matches the specified
        criteria to occur.

        For customization of the callback, use keyword args labeled
        'callback', 'args', and 'kwargs' for the callback fn, callback args,
        and callback kwargs, respectively.  Also note that on return,
        the observer-returned UI element will be included in the first
        argument if 'args' are given.  Note also that if the UI element is
        destroyed, callback should not use it, otherwise the function will
        hang.
        """
        return Observer(self).wait_for(
            filter_=AXCallbacks.match_filter(**kwargs),
            notification=notification,
            timeout=timeout,
        )

    def wait_for_creation(self, timeout=10, notification="AXCreated"):
        """Convenience method to wait for creation of some UI element.

        Returns: The element created
        """
        return self.wait_for(timeout, notification)

    def wait_for_window_to_appear(self, win_name, timeout=10):
        """Convenience method to wait for a window with the given name to
        appear.

        Returns: Boolean
        """
        return self.wait_for(timeout, "AXWindowCreated", AXTitle=win_name)

    def wait_for_window_to_disappear(self, win_name, timeout=10):
        """Convenience method to wait for a window with the given name to
        disappear.

        Returns: Boolean
        """
        # For some reason for the AXUIElementDestroyed notification to fire,
        # we need to have a reference to it first
        return self.wait_for(
            timeout, "AXUIElementDestroyed", AXRole="AXWindow", AXTitle=win_name
        )

    def wait_for_sheet_to_appear(self, timeout=10):
        """Convenience method to wait for a sheet to appear.

        Returns: the sheet that appeared (element) or None
        """
        return self.wait_for_creation(timeout, "AXSheetCreated")

    def wait_for_value_to_change(self, timeout=10):
        """Convenience method to wait for value attribute of given element to
        change.

        Some types of elements (e.g. menu items) have their titles change,
        so this will not work for those.  This seems to work best if you set
        the notification at the application level.

        Returns: Element or None
        """
        # Want to identify that the element whose value changes matches this
        # object's.  Unique identifiers considered include role and position
        # This seems to work best if you set the notification at the application
        # level
        return self.wait_for(timeout, "AXValueChanged")

    def wait_for_focus_to_change(self, new_focused_element, timeout=10):
        """Convenience method to wait for focused element to change (to new
        element given).

        Returns: Boolean
        """
        return self.wait_for(
            timeout,
            "AXFocusedUIElementChanged",
            AXRole=new_focused_element.AXRole,
            AXPosition=new_focused_element.AXPosition,
        )

    def wait_for_focused_window_to_change(self, next_win_name, timeout=10):
        """Convenience method to wait for focused window to change

        Returns: Boolean
        """
        return self.wait_for(timeout, "AXFocusedWindowChanged", AXTitle=next_win_name)

    def wait_for_focus_to_match_criteria(self, timeout=10, **kwargs):
        """Convenience method to wait for focused element to change
        (to element matching kwargs criteria).

        Returns: Element or None

        """
        return self.wait_for(timeout, "AXFocusedUIElementChanged", **kwargs)



import fnmatch


def match_filter(**kwargs):
    def _match(obj):
        for k in kwargs.keys():
            try:
                val = getattr(obj, k)
            except AttributeError:
                return False
            if isinstance(val, str):
                if not fnmatch.fnmatch(val, kwargs[k]):
                    return False
            else:
                if val != kwargs[k]:
                    return False
        return True

    return _match



import logging
import signal
import threading
import time

from ApplicationServices import AXObserverGetRunLoopSource, NSDefaultRunLoopMode
from macuiauto._macos import (
    PAXObserverAddNotification,
    PAXObserverCallback,
    PAXObserverCreate,
    PAXObserverRemoveNotification,
)
from CoreFoundation import CFRunLoopAddSource, CFRunLoopGetCurrent
from PyObjCTools import AppHelper

try:
    from PyObjCTools import MachSignals
except ImportError:

    class MachSignals:
        signal = signal.signal


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _sigHandler(sig):
    AppHelper.stopEventLoop()
    raise KeyboardInterrupt("Keyboard interrupted Run Loop")


class Observer:
    def __init__(self, uielement=None):
        self.ref = uielement
        self.callback = None
        self.callback_result = None

    def wait_for(self, notification=None, filter_=None, timeout=5):
        self.callback_result = None

        @PAXObserverCallback
        def _callback(observer, element, notification, refcon):
            logger.debug("CALLBACK")
            logger.debug("%s, %s, %s, %s" % (observer, element, notification, refcon))
            ret_element = self.ref.__class__(element)
            if filter_(ret_element):
                self.callback_result = ret_element

        observer = PAXObserverCreate(self.ref.pid, _callback)

        PAXObserverAddNotification(
            observer, self.ref.ref, notification, id(self.ref.ref)
        )

        # Add observer source to run loop
        CFRunLoopAddSource(
            CFRunLoopGetCurrent(),
            AXObserverGetRunLoopSource(observer),
            NSDefaultRunLoopMode,
        )

        def event_stopper():
            end_time = time.time() + timeout
            while time.time() < end_time:
                if self.callback_result is not None:
                    break
            AppHelper.callAfter(AppHelper.stopEventLoop)

        event_watcher = threading.Thread(target=event_stopper)
        event_watcher.daemon = True
        event_watcher.start()

        # Set the signal handlers prior to running the run loop
        oldSigIntHandler = MachSignals.signal(signal.SIGINT, _sigHandler)
        AppHelper.runConsoleEventLoop()
        MachSignals.signal(signal.SIGINT, oldSigIntHandler)

        PAXObserverRemoveNotification(observer, self.ref.ref, notification)

        return self.callback_result



"""
Wrap objc calls to raise python exception
"""
# flake8: noqa: B950
from ApplicationServices import (
    AXObserverAddNotification,
    AXObserverCreate,
    AXObserverRemoveNotification,
    AXUIElementCopyActionNames,
    AXUIElementCopyAttributeNames,
    AXUIElementCopyAttributeValue,
    AXUIElementCopyElementAtPosition,
    AXUIElementGetPid,
    AXUIElementIsAttributeSettable,
    AXUIElementPerformAction,
    AXUIElementSetAttributeValue,
    AXUIElementSetMessagingTimeout,
)
from macuiauto import errors
from objc import callbackFor

PAXObserverCallback = callbackFor(AXObserverCreate)


def PAXObserverCreate(application, callback):
    """
    Creates a new observer that can receive notifications
    from the specified application.

    Args:
        application: The process ID of the application
        callback: The callback function

    Returns: an AXObserverRef representing the observer object

    """
    error_code, observer = AXObserverCreate(application, callback, None)
    error_messages = {
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value",
        errors.kAXErrorFailure: "There is some sort of system memory failure",
    }
    errors.check_ax_error(error_code, error_messages)
    return observer


def PAXObserverAddNotification(observer, element, notification, refcon):
    """
    Registers the specified observer to receive notifications from
    the specified accessibility object

    Args:
        observer: The observer object created from a call to AXObserverCreate
        element: The accessibility object for which to observe notifications
        notification: The name of the notification to observe
        refcon: Application-defined data passed to the callback when it is called

    """
    error_code = AXObserverAddNotification(observer, element, notification, refcon)
    error_messages = {
        errors.kAXErrorInvalidUIElementObserver: "The observer is not a valid AXObserverRef type.",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value or the length of the notification name is greater than 1024.",
        errors.kAXErrorNotificationUnsupported: "The accessibility object does not support notifications (note that the system-wide accessibility object does not support notifications).",
        errors.kAXErrorNotificationAlreadyRegistered: "The notification has already been registered.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorFailure: "There is some sort of system memory failure.",
    }
    errors.check_ax_error(error_code, error_messages)


def PAXObserverRemoveNotification(observer, element, notification):
    """
    Removes the specified notification from the list of notifications the
    observer wants to receive from the accessibility object.

    Args:
        observer: The observer object created from a call to AXObserverCreate
        element: The accessibility object for which this observer observes notifications
        notification: The name of the notification to remove from
            the list of observed notifications

    """
    error_code = AXObserverRemoveNotification(observer, element, notification)
    error_messages = {
        errors.kAXErrorInvalidUIElementObserver: "The observer is not a valid AXObserverRef type.",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value or the length of the notification name is greater than 1024.",
        errors.kAXErrorNotificationUnsupported: "The accessibility object does not support notifications (note that the system-wide accessibility object does not support notifications).",
        errors.kAXErrorNotificationNotRegistered: "This observer has not registered for any notifications.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorFailure: "There is some sort of system memory failure.",
    }
    errors.check_ax_error(error_code, error_messages)


def PAXUIElementCopyAttributeValue(element, attribute):
    """
    Returns the value of an accessibility object's attribute

    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name

    Returns: the value associated with the specified attribute

    """
    error_code, attrValue = AXUIElementCopyAttributeValue(element, attribute, None)
    error_messages = {
        errors.kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        errors.kAXErrorNoValue: "The specified attribute does not have a value.",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)
    return attrValue


def PAXUIElementIsAttributeSettable(element, attribute):
    """
    Returns whether the specified accessibility object's attribute can be modified

    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name

    Returns: a Boolean value indicating whether the attribute is settable

    """
    error_code, settable = AXUIElementIsAttributeSettable(element, attribute, None)
    error_messages = {
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way (often due to a timeout).",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        errors.kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        errors.kAXErrorNoValue: "The specified attribute does not have a value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)
    return settable


def PAXUIElementSetAttributeValue(element, attribute, value):
    """
    Sets the accessibility object's attribute to the specified value

    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name
        value: The new value for the attribute

    """
    error_code = AXUIElementSetAttributeValue(element, attribute, value)
    error_messages = {
        errors.kAXErrorIllegalArgument: "The value is not recognized by the accessible application or one of the other arguments is an illegal value.",
        errors.kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)


def PAXUIElementCopyAttributeNames(element):
    """
    Returns a list of all the attributes supported by the specified accessibility object

    Args:
        element: The AXUIElementRef representing the accessibility object

    Returns: an array containing the accessibility object's attribute names

    """
    error_code, names = AXUIElementCopyAttributeNames(element, None)
    error_messages = {
        errors.kAXErrorAttributeUnsupported: "The specified AXUIElementRef does not support the specified attribute.",
        errors.kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorFailure: "There was a system memory failure.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)
    return names


def PAXUIElementCopyActionNames(element):
    """
    Returns a list of all the actions the specified accessibility object can perform

    Args:
        element: The AXUIElementRef representing the accessibility object

    Returns: an array of actions the accessibility object can perform
        (empty if the accessibility object supports no actions)

    """
    error_code, names = AXUIElementCopyActionNames(element, None)
    error_messages = {
        errors.kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorFailure: "There was some sort of system memory failure.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)
    return names


def PAXUIElementPerformAction(element, action):
    """
    Requests that the specified accessibility object perform the specified action

    Args:
        element: The AXUIElementRef representing the accessibility object
        action: The action to be performed

    """
    error_code = AXUIElementPerformAction(element, action)
    error_messages = {
        errors.kAXErrorActionUnsupported: "The specified AXUIElementRef does not support the specified action (you will also receive this error if you pass in the system-wide accessibility object).",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way or the application has not yet responded.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)


def PAXUIElementGetPid(element):
    """
    Returns the process ID associated with the specified accessibility object

    Args:
        element: The AXUIElementRef representing an accessibility object

    Returns: the process ID associated with the specified accessibility object

    """
    error_code, pid = AXUIElementGetPid(element, None)
    error_messages = {
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
    }
    errors.check_ax_error(error_code, error_messages)
    return pid


def PAXUIElementCopyElementAtPosition(application, x, y):
    """
    Returns the accessibility object at the specified position in
    top-left relative screen coordinates

    Args:
        application: The AXUIElementRef representing the application that
            contains the screen coordinates (or the system-wide accessibility object)
        x: The horizontal position
        y: The vertical position

    Returns: the accessibility object at the position specified by x and y

    """
    error_code, element = AXUIElementCopyElementAtPosition(application, x, y, None)
    error_messages = {
        errors.kAXErrorNoValue: "There is no accessibility object at the specified position.",
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        errors.kAXErrorCannotComplete: "The function cannot complete because messaging has failed in some way.",
        errors.kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    errors.check_ax_error(error_code, error_messages)
    return element


def PAXUIElementSetMessagingTimeout(element, timeoutInSeconds):
    """
    Sets the timeout value used in the accessibility API

    Args:
        element: The AXUIElementRef representing an accessibility object
        timeoutInSeconds: The number of seconds for the new timeout value

    """
    error_code = AXUIElementSetMessagingTimeout(element, timeoutInSeconds)
    error_messages = {
        errors.kAXErrorIllegalArgument: "One or more of the arguments is an illegal value (timeout values must be positive).",
        errors.kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
    }
    errors.check_ax_error(error_code, error_messages)



from ApplicationServices import (
    kAXErrorActionUnsupported,
    kAXErrorAPIDisabled,
    kAXErrorAttributeUnsupported,
    kAXErrorCannotComplete,
    kAXErrorFailure,
    kAXErrorIllegalArgument,
    kAXErrorInvalidUIElement,
    kAXErrorInvalidUIElementObserver,
    kAXErrorNotificationAlreadyRegistered,
    kAXErrorNotificationNotRegistered,
    kAXErrorNotificationUnsupported,
    kAXErrorNotImplemented,
    kAXErrorNoValue,
    kAXErrorSuccess,
)


class AXError(Exception):
    pass


class AXErrorUnsupported(AXError):
    pass


class AXErrorAPIDisabled(AXError):
    pass


class AXErrorInvalidUIElement(AXError):
    pass


class AXErrorCannotComplete(AXError):
    pass


class AXErrorNotImplemented(AXError):
    pass


class AXErrorIllegalArgument(AXError):
    pass


class AXErrorActionUnsupported(AXError):
    pass


class AXErrorNoValue(AXError):
    pass


class AXErrorFailure(AXError):
    pass


class AXErrorInvalidUIElementObserver(AXError):
    pass


class AXErrorNotificationUnsupported(AXError):
    pass


class AXErrorNotificationAlreadyRegistered(AXError):
    pass


class AXErrorNotificationNotRegistered(AXError):
    pass


class AXErrorAttributeUnsupported(AXError):
    pass


def AXErrorFactory(code):
    return {
        kAXErrorAPIDisabled: AXErrorAPIDisabled,
        kAXErrorInvalidUIElement: AXErrorInvalidUIElement,
        kAXErrorCannotComplete: AXErrorCannotComplete,
        kAXErrorNotImplemented: AXErrorNotImplemented,
        kAXErrorIllegalArgument: AXErrorIllegalArgument,
        kAXErrorNoValue: AXErrorNoValue,
        kAXErrorFailure: AXErrorFailure,
        kAXErrorInvalidUIElementObserver: AXErrorInvalidUIElementObserver,
        kAXErrorNotificationUnsupported: AXErrorNotificationUnsupported,
        kAXErrorNotificationAlreadyRegistered: AXErrorNotificationAlreadyRegistered,
        kAXErrorNotificationNotRegistered: AXErrorNotificationNotRegistered,
        kAXErrorAttributeUnsupported: AXErrorAttributeUnsupported,
        kAXErrorActionUnsupported: AXErrorActionUnsupported,
    }.get(code, AXErrorUnsupported)


def check_ax_error(error_code, error_messages):
    """
    Returns if code is kAXErrorSuccess.
    Raises an error with given message based on given error code.
    Defaults to AXErrorUnsupported for unknown codes.

    Args:
        error_code: the error code
        error_messages: mapping from error code to error message
    """
    if error_code == kAXErrorSuccess:
        return

    try:
        error_message = error_messages[error_code]
    except KeyError:
        if error_code == kAXErrorFailure:
            error_message = "There is some sort of system memory failure"
        else:
            error_message = "Unknown AX Error: %s" % error_code

    raise AXErrorFactory(error_code)(error_message)



import re
from typing import Any, NamedTuple

from ApplicationServices import (
    AXUIElementGetTypeID,
    AXValueGetType,
    NSPointFromString,
    NSRangeFromString,
    NSRectFromString,
    NSSizeFromString,
    kAXValueCFRangeType,
    kAXValueCGPointType,
    kAXValueCGRectType,
    kAXValueCGSizeType,
)
from CoreFoundation import CFArrayGetTypeID, CFGetTypeID, CFStringGetTypeID


class CGRect(NamedTuple):
    x: float
    y: float
    width: float
    height: float


class CFRange(NamedTuple):
    location: float
    length: int


class CGPoint(NamedTuple):
    x: float
    y: float


class CGSize(NamedTuple):
    width: float
    height: float


class Converter:
    def __init__(self, axuielementclass=None):
        self.app_ref_class = axuielementclass

    def convert_value(self, value):
        if CFGetTypeID(value) == CFStringGetTypeID():
            try:
                return str(value)
            except UnicodeEncodeError:
                return str(value.encode("utf-8"))
        if CFGetTypeID(value) == AXUIElementGetTypeID():
            return self.convert_app_ref(value)
        if CFGetTypeID(value) == CFArrayGetTypeID():
            return self.convert_list(value)
        if AXValueGetType(value) == kAXValueCGSizeType:
            return self.convert_size(value)
        if AXValueGetType(value) == kAXValueCGPointType:
            return self.convert_point(value)
        if AXValueGetType(value) == kAXValueCFRangeType:
            return self.convert_range(value)
        if AXValueGetType(value) == kAXValueCGRectType:
            return self.convert_rect(value)
        else:
            return value

    def convert_list(self, value):
        return [self.convert_value(item) for item in value]

    def convert_app_ref(self, value):
        return self.app_ref_class(ref=value)

    @staticmethod
    def convert_size(value) -> CGSize:
        repr_searched = re.search("{.*}", str(value)).group()
        # CGSize = namedtuple("CGSize", ["width", "height"])
        size = NSSizeFromString(repr_searched)

        return CGSize(size.width, size.height)

    @staticmethod
    def convert_point(value) -> CGPoint:
        repr_searched = re.search("{.*}", str(value)).group()
        # CGPoint = namedtuple("CGPoint", ["x", "y"])
        point = NSPointFromString(repr_searched)

        return CGPoint(point.x, point.y)

    @staticmethod
    def convert_range(value) -> CFRange:
        repr_searched = re.search("{.*}", str(value)).group()
        # CFRange = namedtuple("CFRange", ["location", "length"])
        range = NSRangeFromString(repr_searched)

        return CFRange(range.location, range.length)

    @staticmethod
    def convert_rect(value: Any) -> CGRect:
        repr_searched = re.search("{.*}", str(value)).group()
        # CGRect = namedtuple("CGRect", ["x", "y", "width", "height"])
        rect = NSRectFromString(repr_searched)

        return CGRect(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height)
